
#include"affichage.h"
#include<cstdlib>
#include<vector>
#include<iostream>
#include <queue>
#include <cmath>

using namespace std;

/* Structure Arbre binaire */
#ifndef STRUCT_ARBREB
#define STRUCT_ARBREB
struct ArbreB{
    ArbreB(string et) {etiquette=et; fg=NULL; fd=NULL;}
    ArbreB(string et, ArbreB* filsGauche, ArbreB* filsDroit) {etiquette=et; fg=filsGauche; fd=filsDroit;}
    string etiquette;
    ArbreB* fg; /* pointeur vers le fils gauche; NULL si vide */
    ArbreB* fd; /* pointeur vers le fils droit; NULL si vide */
};
#endif

/* Le but de ce module est de fournir une fonction d'affichage d'un arbre bianire, en suivant l'exemple suivant:
 
     r
  u-' `-v
 / \     \
a   b     d
 
 En particulier: si un sous-arbre est vide, on réserve quand même la place d'un arbre binaire complet de la bonne hauteur (illustation: fils gauche de v est vide). */


#define CHAR_FILS_GAUCHE "'"
#define CHAR_FILS_DROIT "`"
#define CHAR_FEUILLE_GAUCHE "/"
#define CHAR_FEUILLE_DROITE "\\"



/* imprime k espaces */
void espaces(int k)
{
    int i;
    for (i=0; i<k; i++)
    {
        cout << " ";
    }
}

/* imprime k tirets */
void tirets(int k)
{
    int i;
    for (i=0; i<k; i++)
    {
        cout << "-";
    }
}

/* calcule la hauteur d'un arbre binaire complet de hauteur n, en supposant que toutes les étiquettes ont longueur 1 (appelons cela un arbre binaire complet standard) */
int largeur_par_defaut_selon_hauteur(int n)
{
    if (n==0)
    {
        return 2;
    }
    if (n==1)
    {
        return 5;
        /* 5 car on considère l'arbre suivant:
           p
          / \
         a   b */
    }
    else
    {
        return(2*largeur_par_defaut_selon_hauteur(n-1)+1);
        /* largeur du fils gauche + 1 pour l'étiquette + largeur du fils droit:
              r
          g--' `--d
         / \     / \
        /   \   /   \
        -----   -----    */
    }
}

/* calcule la largeur de l'arbre A, sachant à quel étage on est dans l'arbre total T que l'on veut afficher.
 Numérotage des étages: plus basses feuilles de T -> etage 0 ; racine de T -> étage=profondeur(T) */
int largeur(ArbreB* A, int etage)
{
    if (A==NULL)
    {
        return largeur_par_defaut_selon_hauteur(etage);
        /* si un sous-arbre est vide, on réserve quand même la place d'un sous-arbre binaire complet standard (= toutes les étiquettes ont longueur 1) de hauteur = etage */
    }
    else if (etage==0)
    {
        return(1+(A->etiquette).length());
        /* on compte l'étiquette + 1 pour le caractère "CHAR_FEUILLE_GAUCHE" ou "CHAR_FEUILLE_DROITE":
         
            /   ou  \
         bla         bla       */
    }
    else
        
        return(largeur(A->fg, etage-1)+(A->etiquette).length()+largeur(A->fd, etage-1));
        /* largeur du fils gauche + longeur etiquette + largeur du fils droit */
}

/* Renvoie un pointeur vers le sous-arbre de A désigné par (etages_a_descendre, numero) où
 - etages_a_descendre indique le distance entre la racine de A et le noeud cherché
 - numéro donne le numéro du noeud dans son étage. Autrement dit, numero fonctionne ainsi: tous les noeuds du même étage i sont numérotés de 1 à 2^i de gauche à droite (comme si l'arbre était complet). Autrement dit, le noeud avec (etages_a_descendre=i, numero=j) a pour fils (éventuellement vides) les noeuds (i+1, 2j-1) et (i+1, 2j).
 Si le noeud (i,j) n'existe pas, on renvoie NULL */
ArbreB* sous_arbre(int etages_a_descendre, int numero, ArbreB* A)
{
    if (etages_a_descendre==0)
    {
        /* on a trouvé*/
        return A;
    }
    else
    {
        if (numero<= pow(2, etages_a_descendre-1))
        {
            /* dans ce cas, il faut explorer dans le fils gauche */
            if ((A->fg)!=NULL)
            {
                return(sous_arbre(etages_a_descendre-1, numero, A->fg));
            }
            else
                return NULL;
        }
        else
        {
            /* dans ce cas, il faut explorer dans le fils droit */
            if ((A->fd)!=NULL)
            {
                return(sous_arbre(etages_a_descendre-1, numero-pow(2, etages_a_descendre-1), A->fd));
            }
            else
                return NULL;
            
        }
    }
    
}


/* Renvoie un pointeur vers le père dans l'arbre A (de profondeur prof) du noeud désigné par (i,j):
 - i désigne le numéro de l'étage (i=0 est l'étage des feuilles les plus basses, i=profondeur(A) est l'étage de la racine de A)
 - j désigne le numéro du noeud dans son étage, cf explication ci-dessus */
ArbreB* pere(int i, int j, int prof, ArbreB* A)
{
    if (i==prof)
    {
        return NULL;
    }
    else if (j%2==1)
    {
        return(sous_arbre(prof-i-1, j/2+1, A));
    }
    else
    {
        return(sous_arbre(prof-i, j/2, A));
    }
    
}

/* espaces_ancetres: calcule et imprime la bonne quantité d'espaces nécessaires après l'écriture de l'étiquette du noeud (i,j), en fonction des ancetres (avec i=etage du noeud et j=numero du noeud dans l'étage, cf ci-dessus). On travaille dans un arbre A de profondeur prof.
 En effet:
 - après l'écriture d'un noeud de numéro impair (qui est donc le fils gauche de son père), il faut laisser autant d'espace que la longueur de l'étiquette du père
 - après l'écriture d'un noeud de numéro pair (qui est donc le fils droit de son père), on termine l'écriture d'un sous-arbre enraciné en n certain noeud z. Il faut donc laisser autant d'espaces que la longueur de l'étiquette du père de z.
 Exemple (il faut calculer les espaces marqués ici par des "_" :
                   pere
   noeud impair---'____

 ou 
                          ancetre
           c-------------'
      a---' `---b
   1-'  `-2  3-' `-4 pair________
 */
 
void espaces_ancetre(int i, int j, int prof, ArbreB* A)
{
    ArbreB* p;
    if (i==prof)
    {
        espaces((A->etiquette).length());
    }
    else if (A==NULL)
    {
        espaces(1);
    }
    else
    {
        if (j%2==1) // si j est impair
        {
            // l'ancetre a considérer est le pere de j
            p=pere(i,j, prof, A);
            
            if (p!=NULL)
            {
                espaces((p->etiquette).length());
            }
            else
            {
                // si ce pere est vide, on compte par defaut un seul espace
                espaces(1);
            
            }
            
        }
        else
        {
            // si j est impair, il faut remonter de pere en pere jusqu'à trouver un noeud de numero impair dans son etage.
            espaces_ancetre(i+1,j/2,prof, A);
        }
    }
}

/* fonction affiche_etage: arguments
 - ArbreB* A: arbre dans lequel on travaille
 - prof=profondeur de l'arbre A
 - int i=etage avec: i=profondeur est l'étage de la racine; i=0 est l'étage des feuilles les plus basses (pour i=0, on a deux demi-etages: 'h' pour haut et 'b' pour bas, car, à cause des choix d'affichage, l'affichage des feuilles se fait sur deux lignes: 'h' pour les CHAR_FEUILLES_Gauche/Droite ("/" et "\"), et un demi-etage bas pour les étiquettes des feuilles).
 - demi_etage: 'h' pour haut ou 'b' pour bas si i=0; n'importe quel char sinon.
 - nbNoeuds=2 puissance (prof-i): nombre de noeuds sur l'étage i, comme si l'arbre était complet (s'il ne l'est pas, certains noeuds sont vides). Les noeuds d'un même étage son numérotés de 1 à nbNoeuds, de gauche à droite.
 */
void affiche_etage(int i, ArbreB* A, int prof, int nbNoeuds, char demi_etage)
{
    int j;
    ArbreB* arbre_courant=NULL;
    // on va écrire la ligne de l'étage i (ou du demi-étage, le cas échéant), donc on parcourt tous les noeuds de cet étage de gauche à droite
    for (j=1; j<=nbNoeuds; j++)
    {
        // on trouve l'arbre que l'on va écrire à cet étape: etages_a_descendre=prof-i, numero=j
        arbre_courant=sous_arbre(prof-i, j, A);
        if (arbre_courant!=NULL)
        {
            
            if ((j%2)==1) // si le numero du noeud dans son étage est impair
            {
                if ((i==0)&& (demi_etage=='h')) // demi-étage haut de l'étage 0
                {
                    espaces(largeur(arbre_courant, i)-1);
                    cout << CHAR_FEUILLE_GAUCHE;
                    /*imprime:  (espaces correspondant à la feuille) / */
                }
                else if ((i==0)&& (demi_etage=='b')) //demi-étage bas de l'étage 0
                {
                    cout << arbre_courant->etiquette;
                    espaces(1);
                    /* imprime: etiquette_de_la_feuille + un espace */
                }
                else
                {
                    /* imprime d'abord:
                     (espaces pour sous-arbre gauche) etiquette */
                    espaces(largeur(arbre_courant->fg, i-1));
                    cout << arbre_courant->etiquette;
                    if (i!= prof) // puis, si le noeud courant n'est pas la racine, il s'agit d'un fils gauche, donc il faut imprimer les tirets pour le rattacher à son père. Le nombre de tirets est la largeur du fils droit -1
                    {
                        tirets(largeur(arbre_courant->fd, i-1)-1);
                        cout << CHAR_FILS_GAUCHE;
                        /* imprime: --------' */
                    }
                    /* Total: on a imprimé      (espaces)       etiquette--------'    */
                }
                
            }
            else // sinon, le numero j du noeud dans son etage est pair
            {
                if ((i==0) && (demi_etage=='h'))// demi-étage haut de l'étage 0
                {
                    cout << CHAR_FEUILLE_DROITE ;
                    espaces(largeur(arbre_courant, i)-1);
                     /*imprime:  \ (espaces correspondant à la feuille) */
                }
                else if ((i==0) && (demi_etage=='b')) //demi-étage bas de l'étage 0
                {
                    espaces(1);
                    cout << arbre_courant->etiquette;
                    /* imprime: un espace + etiquette_de_la_feuille */

                }
                else
                {
                    /* imprime:   `------ etiquette  (espaces) */
                    // avec: nombre de tirets= largeur du fils gauche -1
                    // et nombre d'espaces= largeur du fils droit
                    cout << CHAR_FILS_DROIT ;
                    tirets(largeur(arbre_courant->fg, i-1)-1);
                    cout << arbre_courant->etiquette;
                    espaces(largeur(arbre_courant->fd, i-1));
                }
                
            }
        }
        else
        {
            // si le noeud courant est vide, on réserve avec des espaces la place d'un sous-arbre "phantome" binaire complet stantard (= toutes les étiquettes ont longueur 1) de hauteur i
            espaces(largeur_par_defaut_selon_hauteur(i));
        }
        // on imprime maintenant des espaces pour se placer au bon endroit pour commencer le prochain noeud: on appelle espaces_ancetres.
        espaces_ancetre(i,j,prof, A);
        
    }
    cout << endl; // on termine la ligne de cet étage (ou ce demi-étage)
}

/* calcule la profondeur de l'arbre A */
int profondeur_de (ArbreB* A)
{
    if (A==NULL)
    {
        return -1;
    }
    if ((A->fg==NULL)&& (A->fd==NULL))
    {
        return 0;
    }
    else
        return (1+max(profondeur_de(A->fg), profondeur_de(A->fd)));
}


/* Affiche l'arbre A, en commencant par une phrase d'intro annoncant le nom (string nom, en paramètre) de l'arbre */
void affiche_arbre(ArbreB* A, string nom)
{
    cout << "Affichage de l'arbre: " << nom << endl;
    int n=profondeur_de(A);
    int i,j;
    int nbNoeuds=1;
    if (A!=NULL)
    {
        if ((n==0)&&(A!=NULL)) // si A est seulement une feuille, il suffit d'écrire son étiquette
        {
            cout << A->etiquette << endl;
        }
        else
        {
            // sinon, on va afficher les étages un par un, de i=n (tout en haut, étage de la racine) à i=1  (avant-dernier étage), puis en traitant l'étage 0 (etage des feuilles les plus basses) par demi-etage haut puis bas.
            for (i=n; i>0; i--)
            {
                affiche_etage(i, A, n, nbNoeuds,'n');
                nbNoeuds=2*nbNoeuds;
            }
            affiche_etage(0,A,n,nbNoeuds, 'h');
            affiche_etage(0,A,n,nbNoeuds, 'b');
        }
        
    }
    
    cout << endl;
}

